/*
 * thrulay.c -- network throughput tester (the client part).
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 * 
 * Copyright 2003, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <limits.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include "util.h"
#include "tsc.h"
#include "rcs.h"

RCS_ID("@(#) $Id: thrulay.c,v 1.29 2005/03/22 02:51:23 shalunov Exp $")

#define THRULAY_VERSION "thrulay/2"
#define THRULAY_GREET THRULAY_VERSION "+"

#define IP_HEADER_SIZE 20
#define UDP_HEADER_SIZE 8
#define HEADER_SIZE (IP_HEADER_SIZE + UDP_HEADER_SIZE)
#define UDP_PAYLOAD_SIZE 24

#define UDP_PORT		5003
#define TRY_UDP_PORTS		1000

static void
usage(void)
{
	fprintf(stderr,
		"Usage: thrulay [-t#] [-i#] [-w#] [-l#] [-p#] [-u#] host\n"
		"\t-t#\t\ttest duration, in seconds (default: 60s)\n"
		"\t-i#\t\treporting interval, in seconds (default: 1s)\n"
		"\t-w#\t\twindow, in bytes (default: 4194304B)\n"
		"\t-l#\t\tblock size (default: 8192B)\n"
		"\t-p#\t\tserver port (default: 5003)\n"
		"\t-u#[kMGT]\tUDP mode with given rate (default: off)\n"
		"\t\t\tIn UDP mode, rate is in bits per second and can be\n"
		"\t\t\tfollowed by a SI suffix (k for 1000, M for 100000);\n"
		"\t\t\tdefault packet size, 1500, can be changed with -l;\n"
		"\t\t\twindow size becomes the UDP send buffer size;\n"
		"\t\t\treporting interval is ignored.\n"
		"\thost\t\tserver to send test data to (no default)\n");
	exit(1);
}

/* Make sure 0 <= tv.tv_usec < 1000000.  Return 0 if it was normal,
   positive number otherwise. */
int
normalize_tv(struct timeval *tv)
{
	int result = 0;

	while (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
		result++;
	}
	while (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
		result++;
	}
	return result;
}

/* Process new timestamp.  Keeps internal state.  Return 1 if enough
   data has been already sent and processed, 0 otherwise. */
int
new_timestamp(struct timeval *tv, int test_duration, int reporting_interval,
	      int block_size)
{
	static struct timeval first = {0, 0};
	static struct timeval last = {0, 0};
	static struct timeval next = {0, 0};
	static unsigned int blocks_since_first = 0;
	static unsigned int blocks_since_last = 0;
	static double min_rtt_since_first = 1000.0;
	static double min_rtt_since_last = 1000.0;
	static double max_rtt_since_first = -1000.0;
	static double max_rtt_since_last = -1000.0;
	static double tot_rtt_since_first = 0.0;
	static double tot_rtt_since_last = 0.0;

	struct timeval this;
	double relative;
	int enough = 0;

	if (first.tv_sec == 0 && first.tv_usec == 0) {
		/* Processing the very first block.  Initialize. */
		assert(last.tv_sec == 0 && last.tv_usec == 0 &&
		       next.tv_sec == 0 && next.tv_usec == 0);
		gettimeofday(&first, 0);
		normalize_tv(&first);
		last = first;
		next = last;
		next.tv_sec += reporting_interval;
	}

	gettimeofday(&this, 0);
	normalize_tv(&this);
	blocks_since_first++;
	blocks_since_last++;
	relative = time_diff(tv, &this);
	if (relative < 0) {
		error(ERR_WARNING, "negative delay, ignoring block");
		return 0;
	}
	if (min_rtt_since_first > relative)
		min_rtt_since_first = relative;
	if (min_rtt_since_last > relative)
		min_rtt_since_last = relative;
	if (max_rtt_since_first < relative)
		max_rtt_since_first = relative;
	if (max_rtt_since_last < relative)
		max_rtt_since_last = relative;
	tot_rtt_since_first += relative;
	tot_rtt_since_last += relative;

	if (this.tv_sec > next.tv_sec
	    || (this.tv_sec == next.tv_sec
		&& this.tv_usec > next.tv_usec)) {
		/* A reporting interval has ended. */
		double since_first;

		since_first = time_diff(&first, &last);
		relative = time_diff(&last, &this);
		/* Time must be monotonically increasing, at least
                   roughly, for this program to work. */
		printf("%9.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n", since_first,
		       since_first + relative,
		       (double)blocks_since_last * (double)block_size * 8.0
		       / 1000000.0 / relative,
		       min_rtt_since_last*1000.0,
		       tot_rtt_since_last*1000.0/(double)blocks_since_last,
		       max_rtt_since_last*1000.0);
		fflush(stdout);
		blocks_since_last = 0;
		min_rtt_since_last = 1000.0;
		max_rtt_since_last = -1000.0;
		tot_rtt_since_last = 0.0;
		last = this;
		while (next.tv_sec < this.tv_sec
		       || (next.tv_sec == this.tv_sec
			   && next.tv_usec <= this.tv_usec))
			next.tv_sec += reporting_interval;
	}
	if (this.tv_sec > first.tv_sec+test_duration
	    || (this.tv_sec == first.tv_sec+test_duration
		&& this.tv_usec > first.tv_usec)) {
		/* Time's up.  We're done. */
		enough = 1;
		relative = time_diff(&first, &this);
		if (relative < 0)
			error(ERR_WARNING, "non-monotonic time, I'm confused");
		printf("#%8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n", 0.0, relative,
		       (double)blocks_since_first * (double)block_size * 8.0
		       / 1000000.0 / relative, min_rtt_since_first*1000.0,
		       tot_rtt_since_first*1000.0/(double)blocks_since_first,
		       max_rtt_since_first*1000.0);
		fflush(stdout);
	}

	return enough;
}

/*
 * Return a socket connected to the right TCP port on the right
 * server.  Use the given window size, if non-zero, locally.  Store
 * the actual window size in *real_window.  Store server socket
 * address structure in *server (if server is non-NULL).
 */
int
name2socket(char *server_name, int port, int window, int *real_window,
	    struct sockaddr_in *server)
{
	/* XXX: This code will need to be rewritten to support IPv6 later. */
	struct hostent *hp = 0;
	struct sockaddr_in sin;
	int s;

	memset(&sin, 0, sizeof sin);
        if (! inet_aton(server_name, &sin.sin_addr)) {
		hp = gethostbyname(server_name);
		if (! hp) {
			herror(server_name);
			error(ERR_FATAL, "could not resolve host");
			/* NOTREACHED */
		}
		sin.sin_family = hp->h_addrtype;
		bcopy(hp->h_addr, (char *)&sin.sin_addr,
		      (size_t) hp->h_length);
	} else
		sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	if ((s = socket(sin.sin_family, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		error(ERR_FATAL, "could not create a socket");
		/* NOTREACHED */
	}
	if (window)
		*real_window = set_window_size(s, window);
 	if (connect(s, (struct sockaddr*) &sin, sizeof sin) == -1) {
		perror("connect");
		error(ERR_FATAL, "could not establish connection to server");
		/* NOTREACHED */
	}
	if (server)
		memcpy(server, &sin, sizeof(sin));
	return s;
}

/* Read the thrulay greeting from a TCP connection associated with socket s. */
void
read_greeting(int s)
{
	char buf[1024];
	int rc;
	size_t greetlen = sizeof(THRULAY_GREET) - 1;

	rc = read_exactly(s, buf, greetlen);
	assert(rc <= (int) greetlen);
	if (rc != (int) greetlen) {
		if (rc == -1)
			perror("read");
		error(ERR_FATAL, "could not read server banner");
		/* NOTREACHED */
	}
	if (strncmp(buf, THRULAY_VERSION, sizeof THRULAY_VERSION - 1) != 0)
		error(ERR_FATAL, "not a thrulay server responded");
	if (buf[greetlen - 1] != '+') {
		error(ERR_WARNING, "connection rejected");
		rc = read(s, buf, sizeof buf - 1);
		buf[sizeof buf - 1] = '\0';
		if (rc == -1) {
			perror("reading rejection reason");
			error(ERR_FATAL, "could not read rejection reason");
			/* NOTREACHED */
		}
		assert(rc < (int) sizeof buf);
		buf[rc] = '\0';
		fprintf(stderr, "server said: %s", buf);
		if (buf[rc-1] != '\n')
			fprintf(stderr, "\n");
		error(ERR_FATAL, "stop");
		/* NOTREACHED */
	}
}

void
write_proposal(int s, char *proposal, int proposal_size)
{
	int rc;

	rc = write_exactly(s, proposal, (size_t) proposal_size);
	assert(rc <= proposal_size);
	if (rc < proposal_size) {
		if (rc == -1)
			perror("write");
		error(ERR_FATAL, "could not write session proposal");
		/* NOTREACHED */
	}
}

/* Read response to a proposal.  Return the size of the response. */
int
read_response(int s, char *buf, int max)
{
	int rc;

	/* XXX: Assume that few-byte session response will come in one
           TCP packet and read in one block. */
	rc = read(s, buf, max - 1);
	assert(rc < max);
	if (rc == -1) {
		perror("read");
		error(ERR_FATAL, "could not read session response");
		/* NOTREACHED */
	} else if (rc == 0) {
		error(ERR_FATAL, "server closed connection after proposal");
		/* NOTREACHED */
	}
	assert(rc > 0);
	buf[rc] = '\0';
	return rc;
}

void
thrulay(char *server_name, int test_duration, int reporting_interval,
	int window, int block_size, int port)
{
	char buf[1024];
	int s;
	int rc;
	int to_write;
	int w, b;	/* Server-specified window and block size. */
	int local_w;	/* Local TCP window, actual. */
	char *block;
	int enough = 0;

	s = name2socket(server_name, port, window, &local_w, NULL);
	assert(s > 0);
	read_greeting(s);
	to_write = snprintf(buf, sizeof buf, "%s:t:%u:%u+",
			    THRULAY_VERSION, window, block_size);
	assert(to_write > 0 && to_write < (int) sizeof buf);
	write_proposal(s, buf, to_write);
	read_response(s, buf, sizeof(buf));
	w = b = -1;
	/* XXX: Very long numbers will not, of course, be processed
           correctly by sscanf() below.  We could, if the number
           exceeds a certain value (e.g., has more than a certain
           number of characters, use some ``large'' -- still supported
           -- value for window or block size.  It's not worth the
           trouble. */
	rc = sscanf(buf, "%u:%u+", &w, &b);
	if (rc != 2)
		error(ERR_FATAL, "malformed session response from server");
	assert(w >= 0 && b >= 0);	/* Unsigned formats. */
	if (b < MIN_BLOCK)
		error(ERR_FATAL, "server gave block size less than MIN_BLOCK");
	if (b > MAX_BLOCK)
		error(ERR_FATAL, "server gave block size too large");
	if (w < 1500)
		error(ERR_WARNING, "server gave ridiculously small window");
	block = malloc((size_t) b);
	if (! block)
		error(ERR_FATAL, "could not allocate memory");

	printf("# local window = %dB; remote window = %dB\n",
	       local_w, w);
	if (block_size == b)
		printf("# block size = %dB\n", b);
	else
		printf("# requested block size = %dB; "
		       "actual block size = %dB\n", block_size, b);
	printf("# test duration = %ds; reporting interval = %ds\n",
	       test_duration, reporting_interval);
	printf("# begin, s  end, s  Mb/s    RTT, ms: min  "
	       "avg      max\n");
	fflush(stdout);

	assert((unsigned int) s < FD_SETSIZE);
	memset(block, '2', (size_t) b); /* Nothing special about '2'. */
	while (! enough) {
		struct timeval tv;

		rc = gettimeofday(&tv, 0);
		if (rc == -1) {
			perror("gettimeofday");
			error(ERR_FATAL, "could not get time of day");
		}
		memcpy(block, &tv, sizeof tv);
		rc = write_exactly(s, block, (size_t) b);
		if (rc == -1) {
			perror("write to remote");
			error(ERR_WARNING, "premature end of test");
			enough = 1; /* FIXME: think */
		}
		do {
			fd_set read_set, write_set, except_set;
			struct timeval zero = {0, 0};

			FD_ZERO(&read_set);
			FD_ZERO(&write_set);
			FD_ZERO(&except_set);
			FD_SET(s, &read_set);
			rc = select(s+1, &read_set, &write_set,
				    &except_set, &zero);
			if (rc > 0) {
				int rc_read;

				assert(rc == 1 && FD_ISSET(s, &read_set));
				rc_read = read_exactly(s, buf, BLOCK_HEADER);
				assert(rc_read <= BLOCK_HEADER);
				if (rc_read == -1) {
					perror("read");
					error(ERR_WARNING,
					      "premature end of test");
					enough = 1;
					rc = -1;
				}
				memcpy(&tv, buf, sizeof tv);
				enough = new_timestamp(&tv, test_duration,
					      reporting_interval,
					      block_size);
			}
		} while (rc > 0 && ! enough);
	}
}

void
thrulay_udp(char *server_name, int test_duration, u_int64_t rate,
	    int udp_buffer_size, int packet_size, int port)
{
	int rc;
	int s;
	char buf[65536];
	char random_state[256];
	int to_write;
	int client_port, server_port;
	int protocol_rate;
	int udp_sock;
	int tries;
	u_int64_t npackets, packet;
	u_int32_t msb, lsb;
	struct sockaddr_in client, server;
	char nonce[8];
	int n, response_size;
	double urand;
	double erand;
	double emean;
	u_int64_t target_tick, ticks, ticks_per_second;

	s = name2socket(server_name, port, 0, NULL, &server);
	assert(s > 0);
	read_greeting(s);
	udp_sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (udp_sock < 0) {
		perror("socket");
		error(ERR_FATAL, "could not get UDP socket");
		/* NOTREACHED */
	}
	/* Get a port number for the client. */
	client_port = UDP_PORT;
	memset(&client, 0, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_addr.s_addr = INADDR_ANY;
	tries = 0;
	do {
		client_port++;
		tries++;
		client.sin_port = htons(client_port);
		rc = bind(udp_sock, (struct sockaddr *)&client,
			  sizeof(client));
	} while ((rc < 0) && (tries < TRY_UDP_PORTS));
	protocol_rate = ((1000/8) * rate)/packet_size;
	npackets = (test_duration*rate)/(8*packet_size);
	to_write = snprintf(buf, sizeof buf, "%s:u:%u:%u:%u:%llu+",
			    THRULAY_VERSION, client_port, packet_size,
			    protocol_rate, npackets);
	write_proposal(s, buf, to_write);
	/*
	 * Both the server and the client do calibration
	 * simultaneously, after proposal is sent.  This allows to
	 * minimize pointless waiting.  Do not move init_tsc() call
	 * around.
	 */
	init_tsc();
	response_size = read_response(s, buf, sizeof(buf));
	if (strcmp(buf, "u:-") == 0)
		error(ERR_FATAL, "server rejected our UDP proposal");
	rc = sscanf(buf, "%u:%u:%u:%llu:%n", &server_port, &packet_size,
		    &protocol_rate, &npackets, &n);
	assert((server_port >= 0) && (packet_size >= 0) &&
	       (protocol_rate >= 0));		/* Unsigned formats. */
	if ((rc != 4) || (response_size != n+9) || (buf[n+8] != '+') ||
	    (packet_size > (int) sizeof(buf))) {
		error(ERR_FATAL, "malformed session response from server");
	}
	memcpy(nonce, buf+n, sizeof(nonce));

	/* Disable keep-alives on the control TCP connection. */
	/* FIXME. */

	/* Increase the buffer space of the sending UDP socket. */
	set_window_size_directed(udp_sock, udp_buffer_size, SO_SNDBUF);

	ticks_per_second = get_tps();
	memset(buf, '2', sizeof(buf));	/* Nothing special about '2'. */
	memcpy(buf, nonce, sizeof(nonce));
	initstate(getpid() + getppid() + time(NULL),
		  random_state, sizeof(random_state));
	emean = 1000.0/(double) protocol_rate;
	target_tick = read_tsc();
	for (packet = 0; packet < npackets; packet++) {
		urand = (random()+1.0)/(double) LONG_MAX;
		assert(urand > 0);
		assert(urand <= 1);
		erand = -emean*log(urand);
		assert(erand >= 0);
		ticks = erand*ticks_per_second;
		target_tick += ticks;
		msb = htonl(packet >> 32);	/* Most significant. */
		lsb = htonl(packet & ((1ULL<<32)-1));/* Least significant. */
		memcpy(buf+8, &msb, 4);
		memcpy(buf+12, &lsb, 4);
		while ((ticks = read_tsc()) < target_tick)
			/* Busy wait. */;
		tsc2ntp(ticks, buf+16);
		rc = sendto(udp_sock, buf, packet_size - HEADER_SIZE, 0,
			    (struct sockaddr *)&server, sizeof(server));
		if (rc == -1) {
			perror("sendto");
			error(ERR_FATAL, "sending UDP packet failed");
		}
	}
	sleep(1);		/* Let the UDP traffic drain. */
	snprintf(buf, sizeof(buf), "+%llu:", npackets);
	rc = write_exactly(s, buf, strlen(buf));
	if (rc == -1)
		error(ERR_FATAL, "could not write terminating message");
	while ((rc = read(s, buf, sizeof(buf) - 1)) != 0) {
		if (rc == -1) {
			perror("read");
			error(ERR_FATAL, "could not read session results");
		}
		write_exactly(1, buf, rc);
	}
}

/* 
 * Convert a string like "12k" to a number like 12000.  Return zero on
 * error.
 */
u_int64_t
rate2i(char *s)
{
	u_int64_t r;	/* Result. */
	char *p;
	int l;
	int suffix = 0;

	/* First, set the multiple. */
	l = strlen(s);
	switch (s[l-1]) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		r = 1;
		break;
	case 'k':
	case 'K':
		r = 1000;
		suffix = 1;
		break;
	case 'm':
	case 'M':
		r = 1000000;
		suffix = 1;
		break;
	case 'g':
	case 'G':
		r = 1000000000ULL;
		suffix = 1;
		break;
	case 't':
	case 'T':
		r = 1000000000000ULL;
		suffix = 1;
		break;
	default:
		r = 0;
	}
	if (suffix)
		s[l-1] = '\0';
	r *= strtoll(s, &p, 10);
	if (!*s || *p)	/* Invalid character found. */
		r = 0;
	return r;
}

int
main(int argc, char *argv[])
{
	char *server_name;		/* Where to send test traffic. */
	int test_duration = 60;		/* For how long, in seconds. */
	int reporting_interval = 1;	/* How often to report progress. */
	int window = 4194304;		/* Window size, in bytes. */
	int block_size;			/* Block size, in bytes. */
	int port = 5003;		/* Server TCP port number. */
	u_int64_t rate = 0;		/* Rate in b/s.  Non-zero means UDP. */
	int ch;

	while ((ch = getopt(argc, argv, "t:i:w:l:p:u:")) != -1)
		switch (ch) {
		case 't':
			test_duration = atoi(optarg);
			if (test_duration <= 0) {
				fprintf(stderr, "test duration must be "
					"a positive integer (in seconds)\n");
				usage();
			}
			break;
		case 'i':
			reporting_interval = atoi(optarg);
			if (reporting_interval <= 0) {
				fprintf(stderr, "reporting interval must be "
					"a positive integer (in seconds)\n");
				usage();
			}
			break;
		case 'w':
			window = atoi(optarg);
			if (window <= 0) {
				fprintf(stderr, "window must be "
					"a positive integer (in bytes)\n");
				usage();
			}
			break;
		case 'l':
			block_size = atoi(optarg);
			if (block_size <= 0) {
				fprintf(stderr, "block size must be "
					"a positive integer (in bytes)\n");
				usage();
			}
			break;
		case 'p':
			port = atoi(optarg);
			if (port <= 0) {
				fprintf(stderr, "port must be "
					"a positive integer\n");
				usage();
			}
			break;
		case 'u':
			rate = rate2i(optarg);
			if (rate == 0) {
				fprintf(stderr, "rate must be positive\n");
				usage();
			}
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();
	server_name = argv[0];
#if 0
	if (rate && reporting_interval) {
		fprintf(stderr, "No reporting interval with UDP tests.\n");
		usage();
	}
#endif
	if (rate)
		block_size = 1500;
	else
		block_size = 8192;

	/* Ignore SIGPIPE.  Always do in any socket code. */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("signal(SIGPIPE, SIG_IGN)");
		error(ERR_FATAL, "could not ignore SIGPIPE");
	}

	assert(sizeof(struct timeval) <= BLOCK_HEADER);

	if (rate)
		thrulay_udp(server_name, test_duration, rate, window,
			    block_size, port);
	else
		thrulay(server_name, test_duration, reporting_interval,
			window, block_size, port);
	exit(0);
	/* NOTREACHED */
}
