/*
 * thrulayd.c -- network throughput tester (the server part).
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 * 
 * Copyright 2003, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/time.h>
#include "util.h"
#include "tsc.h"
#include "assertd.h"
#include "rcs.h"

RCS_ID("@(#) $Id: thrulayd.c,v 1.19 2005/03/22 02:51:23 shalunov Exp $")

/*
 * XXX: Linux has limits.h and an LLONG_MAX definition, but it is not
 * defined unless extra symbols are defined when the file is included.
 * Since the value of the largest signed 64-bit number is not highly
 * volatile, a definition is included right here to fix the bug of not
 * defining this in Linux.
 */
#ifndef LLONG_MAX
#define LLONG_MAX    9223372036854775807LL
#endif

#define LISTEN_BACKLOG		128
#define INDICATOR		"thrulay"
#define THRULAY_VERSION		"2"
#define THRULAY_GREET		"thrulay/2+"
#define UDP_PORT		5003
#define TRY_UDP_PORTS		1000
#define UDP_BUFFER_SIZE		4194304
#define URANDOM_PATH		"/dev/urandom"

void
usage(void)
{
	fprintf(stderr, "Usage: thrulayd [-w#] [-p#]\n"
		"\t-w#\twindow, in bytes (default: 4194304B)\n"
		"\t-p#\tserver port (default: 5003)\n");
	exit(1);
}

void
sighandler(int sig)
{
	int status;

	switch (sig) {
	case SIGCHLD:
		while (waitpid(-1, &status, WNOHANG) > 0)
			;
		break;
	case SIGHUP:
		syslog(LOG_NOTICE, "got SIGHUP, don't know what to do");
		break;
	case SIGPIPE:
		break;
	default:
		syslog(LOG_ALERT, "got signal %d, "
		       "but don't remember intercepting it, aborting", sig);
		abort();
	}
}

void
log_client_address(int fd)
{
	struct sockaddr_in client;
	int len = sizeof client;
	int rc;
  
	rc = getpeername(fd, (struct sockaddr *) &client, &len);
	if (rc == -1) {
		syslog(LOG_ERR, "getpeername: %m, possible port scan");
		_exit(1);
	}
	syslog(LOG_NOTICE, "connection from %s", inet_ntoa(client.sin_addr));
}

void
udp_test(int fd, char* proposal)
{
	int rc;
	int client_port;
	int packet_size;
	int udp_buffer_size;
	u_int64_t rate;		/* In packets per 1000 seconds. */
	u_int64_t npackets;	/* Proposed number of packets. */
	u_int64_t packets_received = 0;	/* Actually received. */
	u_int64_t packets_sent = 0;	/* What the client said at the end. */
	char nonce[8];
	char buffer[65536];
	int response_size;
	int sock;
	int fromlen;
	/* Decrement will be canceled before first use. */
	int server_port = UDP_PORT - 1;
	struct sockaddr_in server;
	struct sockaddr_in from;
	int tries;
	int random_fd;
	int nfds;
	fd_set readfds;
	int test_done = 0;
	int graceful_termination = 0;
	u_int64_t start, end;	/* Start and end of the test. */
	u_int64_t ticks_per_second;
	int64_t min_delay = LLONG_MAX;	/* Minimum delay in (our) ticks. */
	int64_t delay;			/* This packets delay in our ticks. */

	/*
	 * Both the server and the client do calibration
	 * simultaneously, after proposal is sent.  This allows to
	 * minimize pointless waiting.  Do not move init_tsc() call
	 * around.
	 *
	 * In the server, we could have called init_tsc() in main(),
	 * but that creates large errors as the server ages.
	 */
	init_tsc();
	rc = sscanf(proposal, "%u:%u:%llu:%llu+", &client_port, &packet_size,
		    &rate, &npackets);
	if (rc != 4) {
		syslog(LOG_WARNING, "malformed parameters in UDP proposal");
		goto log;
	}
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		syslog(LOG_WARNING, "could not obtain UDP socket");
		goto log;
	}
	/* Get a port number for the server. */
	memset(&server, 0, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	tries = 0;
	do {
		server_port++;
		tries++;
		server.sin_port = htons(server_port);
		rc = bind(sock, (struct sockaddr *)&server, sizeof(server));
	} while ((rc < 0) && (tries < TRY_UDP_PORTS));
	if (rc < 0) {
		syslog(LOG_WARNING, "could not bind a UDP socket");
		goto log;
	}
	/* Obtain random data for the nonce. */
	random_fd = open(URANDOM_PATH, O_RDONLY);
	if (random_fd < 0) {
		syslog(LOG_WARNING, "could not open random device");
		goto log;
	}
	rc = read(random_fd, nonce, sizeof(nonce));
	if (rc != sizeof(nonce)) {
		syslog(LOG_WARNING, "could not read urandom for UDP nonce");
		goto log;
	}
	close(random_fd);
	/* 
	 * Build a UDP session response.  Note that since nonce may
	 * contain NUL characters, it is not a C string.  The length
	 * is kept in response_size.
	 */
	sprintf(buffer, "%u:%u:%llu:%llu:", server_port, packet_size,
		rate, npackets);
	response_size = strlen(buffer);
	memcpy(buffer + response_size, nonce, sizeof(nonce));
	response_size += sizeof(nonce);
	buffer[response_size] = '+';
	response_size++;

	/* Disable keep-alives on the control TCP connection. */
	/* FIXME. */

	/* Increase the buffer space of the receiving UDP socket. */
	udp_buffer_size = set_window_size_directed(sock, UDP_BUFFER_SIZE,
						   SO_RCVBUF);

	write_exactly(fd, buffer, response_size);

	/* Set sockets to nonblocking mode: will use select(). */
	/* FIXME. */

	start = read_tsc();
	ticks_per_second = get_tps();
	/* 
	 * Could have set nfds to FD_SETSIZE, but this is marginally
	 * more efficient.
	 */
	nfds = (fd > sock? fd: sock) + 1;
	while (! test_done) {
		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		FD_SET(sock, &readfds);
		rc = select(nfds, &readfds, 0, 0, 0);
		if (FD_ISSET(sock, &readfds)) {
			/* A new UDP packet arrived. */
			/* XXX: Need to loop in case of coalescence. */
			rc = recvfrom(sock, buffer, sizeof(buffer), 0,
				      (struct sockaddr *)&from, &fromlen);
			if (memcmp(buffer, nonce, sizeof(nonce)) == 0) {
				/* Legitimate packet -- nonce matches. */
				packets_received++;
				delay = read_tsc() - ntp2tsc(buffer+16);
				if (delay < min_delay)
					min_delay = delay;
			} else {
				syslog(LOG_WARNING, "nonce does not match");
			}
		}
		if (FD_ISSET(fd, &readfds)) {
			/* Data on the TCP connection -- test done. */
			test_done = 1;
			/* XXX: Assume it comes through in one piece. */
			rc = read(fd, buffer, sizeof(buffer));
			if (rc < 0) {
				syslog(LOG_WARNING, "could not read TCP "
				      "connection at the end of a UDP test");
				goto log;
			}
			if (buffer[0] == '+') {
				rc = sscanf(buffer+1, "%llu:", &packets_sent);
				if (rc != 1) {
					syslog(LOG_WARNING,
					      "malformed end of test");
					goto log;
				}
				graceful_termination = 1;
			}
		}
	}
	/* 
	 * Done with the UDP test.  Set the TCP socket back to
	 * blocking mode.
	 */
	/* FIXME: fcntl */
	sprintf(buffer,
		"server used UDP buffer size of %d bytes\n"
		"client proposed sending %llu packets\n"
		"client said it sent %llu packets\n"
		"server received %llu packets\n"
		"minimum delay (ignoring clock offset) %.3lfms\n"
		"packet loss %lf%%\n",
		udp_buffer_size,
		npackets, packets_sent, packets_received,
		(1000.0*min_delay)/ticks_per_second,
		packets_sent>packets_received?
		((double)(packets_sent-packets_received)) /
		(double)(packets_sent)*100.0: 0);
	write_exactly(fd, buffer, strlen(buffer));
 log:
	end = read_tsc();
	return;
}

void
serve_client(int fd)
{
	int rc;
	char buffer[1024];
	char *buf;
	char *block;
	int w, b;
	int to_write;
	int blocks = 0;
	struct timeval start = {0, 0};

	buf = buffer;
	rc = write_exactly(fd, THRULAY_GREET, sizeof THRULAY_GREET - 1);
	assert(rc < (int) sizeof THRULAY_GREET);
	if (rc < (int) sizeof THRULAY_GREET - 1) {
		if (rc == -1)
			syslog(LOG_WARNING, "write(greeting): %m");
		syslog(LOG_WARNING, "could not write greeting, exiting");
		goto log;
	}
	/* XXX: Assume that few-byte session proposal will come in one
           TCP packet and read in one block. */
	rc = read(fd, buf, sizeof(buffer) - 1);
	assert(rc < (int) sizeof(buffer));
	if (rc == -1) {
		syslog(LOG_NOTICE, "could not read session proposal");
		goto log;
	}
	assert(rc > 0);
	buf[rc] = '\0';
	w = b = -1;
	rc = memcmp(buf, INDICATOR, sizeof(INDICATOR) - 1);
	if (rc != 0) {
		syslog(LOG_WARNING, "malformed protocol indicator");
		goto log;
	}
	buf += sizeof(INDICATOR) - 1;	/* Skip the indicator. */
	if (*buf != '/') {
		syslog(LOG_WARNING, "protocol indicator not followed by '/'");
		goto log;
	}
	buf++;				/* Skip the slash. */
	rc = memcmp(buf, THRULAY_VERSION, sizeof(THRULAY_VERSION) - 1);
	if (rc != 0) {
		syslog(LOG_WARNING, "malformed protocol version");
		goto log;
	}
	buf += sizeof(THRULAY_VERSION) - 1;	/* Skip the version. */
	if (*buf != ':') {
		syslog(LOG_WARNING, "protocol version not followed by ':'");
		goto log;
	}
	buf++;				/* Skip the colon. */
	if ((buf[0] == 'u') && (buf[1] == ':')) {
		buf += 2;		/* Skip the test type. */
		udp_test(fd, buf);
		return;
	} else if ((buf[0] != 't') || (buf[1] != ':')) {
		syslog(LOG_WARNING, "unknown test proposal type");
		goto log;
	}
	buf += 2;			/* Skip "t:". */
	/* Do TCP test. */
	/* XXX: Very long numbers will not, of course, be processed
           correctly by sscanf() below.  We could, if the number
           exceeds a certain value (e.g., has more than a certain
           number of characters, use some ``large'' -- still supported
           -- value for window or block size.  It's not worth the
           trouble. */
	rc = sscanf(buf, "%u:%u+", &w, &b);
	if (rc != 2) {
		syslog(LOG_WARNING, "malformed session proposal from client");
		goto log;
	}
	assert(w >= 0 && b >= 0);	/* Unsigned formats. */
	if (b < MIN_BLOCK)
		b = MIN_BLOCK;
	if (b > MAX_BLOCK)
		b = MAX_BLOCK;
	if (w < 1500)
		w = 1500;
	block = malloc((size_t) b);
	if (! block) {
		syslog(LOG_ALERT, "could not allocate memory");
		goto log;
	}
	w = set_window_size(fd, w);
	to_write = snprintf(buffer, sizeof(buffer), "%u:%u+", w, b);
	assert(to_write > 0 && to_write < (int) sizeof(buffer));
	rc = write_exactly(fd, buffer, (size_t) to_write);
	assert(rc <= to_write);
	gettimeofday(&start, 0);

	/* And now the testing loop. */
	while ((rc = read_exactly(fd, block, (size_t) b)) == b) {
		blocks++;
		rc = write_exactly(fd, block, BLOCK_HEADER);
		if (rc != BLOCK_HEADER) {
			if (rc == -1)
				syslog(LOG_NOTICE, "write(block_header): %m");
			syslog(LOG_NOTICE, "could not write block header");
			goto log;
		}
	}
	if (rc == -1) {
		if (errno != ECONNRESET)
			syslog(LOG_NOTICE, "while testing: read(): %m");
	} else {
		assert(rc == 0);
	}
 log:
	{
		struct timeval end;
		double bytes, seconds;

		if (start.tv_sec == 0 && start.tv_usec == 0) {
			syslog(LOG_NOTICE, "nothing transferred");
			return;
		}
		gettimeofday(&end, 0);
		seconds = time_diff(&start, &end);
		bytes = (double) b * (double) blocks;
		syslog(LOG_NOTICE, "test duration = %.3fs, "
		       "average throughput = %.3fMb/s",
		       seconds, bytes*8.0/seconds/1000000.0);
	}
}

int
main(int argc, char *argv[])
{
	int window = 4194304;		/* Window size, in bytes. */
	int port = 5003;		/* Server TCP port number. */
	int rc;
	int opt;
	int s;
	struct sockaddr_in sin;
	int sinlen = sizeof sin;
	int ch;

	while ((ch = getopt(argc, argv, "w:p:")) != -1)
		switch (ch) {
		case 'w':
			window = atoi(optarg);
			if (window <= 0) {
				fprintf(stderr, "window must be "
					"a positive integer (in bytes)\n");
				usage();
			}
			break;
		case 'p':
			port = atoi(optarg);
			if (port <= 0) {
				fprintf(stderr, "port must be "
					"a positive integer\n");
				usage();
			}
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 0)
		usage();

	/* Ignore SIGPIPE.  Always do in any socket code. */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("ignoring SIGPIPE");
		error(ERR_FATAL, "could not ignore SIGPIPE");
		/* NOTREACHED */
	}
	{
		struct sigaction sa;

		sa.sa_handler = sighandler;
		sa.sa_flags = 0;
		sigemptyset (&sa.sa_mask);
		sigaction (SIGHUP, &sa, NULL);
		sigaction (SIGALRM, &sa, NULL);
		sigaction (SIGCHLD, &sa, NULL);
	}

	openlog("thrulayd", LOG_PID | LOG_CONS | LOG_NDELAY, LOG_DAEMON);
	rc = daemon(0, 0);
	if (rc == -1) {
		perror("daemon");	/* It could be going to /dev/null... */
		exit(1);
		/* NOTREACHED */
	}

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s == -1) {
		syslog(LOG_ERR, "could not create socket: %m, exiting");
		exit(1);
	}
	opt = 1;
	rc = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*) &opt, sizeof opt);
	if (rc == -1)
		syslog(LOG_WARNING,
		       "setsockopt(SO_REUSEADDR): %m, continuing");
	rc = setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char*) &opt, sizeof opt);
	if (rc == -1)
		syslog(LOG_WARNING,
		       "setsockopt(SO_KEEPALIVE): %m, continuing");

	window = set_window_size(s, window);

	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = INADDR_ANY;
	rc = bind(s, (struct sockaddr*) &sin, sizeof sin);
	if (rc == -1) {
		syslog(LOG_ERR, "bind(%d): %m, exiting", port);
		exit(1);
	}
	rc = listen(s, LISTEN_BACKLOG);
	if (rc == -1) {
		syslog(LOG_ERR, "listen(%d): %m, exiting", LISTEN_BACKLOG);
		exit(1);
	}
	syslog(LOG_NOTICE, "thrulayd started, listening on port %d "
	       "with window size %d", port, window);

	/* Main loop. */
	while (1) {
		int fd, pid;

		fd = accept(s, (struct sockaddr*) &sin, &sinlen);
		if (fd == -1) {
			if (errno != EINTR)
				syslog(LOG_WARNING,
				       "accept(): %m, continuing");
			continue;
		}
		pid = fork();
		switch (pid) {
		case 0:
			close(s);
			log_client_address(fd);
			serve_client(fd);
			closelog();
			_exit(0);
			/* NOTREACHED */
		case -1:
			syslog(LOG_ERR, "fork(): %m, closing connection");
			close(fd);
			break;
		default:
			close(fd);
			break;
		}
	}

	/* NOTREACHED */
}
