/*
 * tsc.c -- deal with the TSC register timestamps.
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 * 
 * Copyright 2005, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#include <netinet/in.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "tsc.h"
#include "assertd.h"
#include "rcs.h"

RCS_ID("@(#) $Id: tsc.c,v 1.12 2005/01/12 02:32:13 shalunov Exp $")

#undef DEBUG

#ifdef DEBUG
#include <stdio.h>
#include <stdlib.h>
#define max(x,y) ((x)>(y)? (x): (y))
#define min(x,y) ((x)<(y)? (x): (y))
#endif

/* 
 * The difference between Unix and NTP epochs, in seconds.
 * Source: http://www.eecis.udel.edu/~mills/y2k.html
 */
#define EPOCH_OFFSET 2208988800ULL

/*
 * How long to sleep, in seconds, during initialization.  Larger
 * values result in longer start-up delays but better calibration.  We
 * need to let the TSC counter run for some time to get ticks per
 * second measurement in a meaningful resolution.  With a 1-second
 * sleep, I get relative error in tps in the 0.0001% (or, to put this
 * differently, 1PPM) range, if everything goes well, that is, there
 * are no unlucky context switches.  Clearly, this is no match for,
 * say, NTP's careful calibration (which requires hours to get right);
 * however, the poor calibration is at least partially mitigated by
 * the fact that the test is likely to last only a relatively short
 * time.
 *
 * XXX: Need to include provisions for recalibration during the test
 * for long tests that require low errors in one-way delay (and,
 * hence, one-way jitter) measurement.  Note that this is of little
 * concern for simple measurements of loss.
 */
#define CALIBRATION_INTERVAL 1

/*
 * After initialization, at time t since epoch, the value of TSC is
 * tae+t*tps.  The epoch here is temporary and corresponds to the
 * start of this software, *not* the NTP epoch, 0h on 1 January 1900.
 */
static u_int64_t tps = 0;	/* TSC ticks per second. */
static u_int64_t tae = 0;	/* TSC ticks at epoch. */

/*
 * This is the offset, in NTP's fractional seconds (2^{-32} seconds
 * each) from the NTP epoch to the epoch as used by this instantiation
 * of the program.
 */
static u_int64_t epoch = 0;

/*
 * Initialize TSC reading routines.  Perform calibration.  Must be
 * called (once) before invocation of other TSC-related functions.
 * There is no great harm in calling init_tsc() another time, but this
 * should be avoided in processes spawned with fork() where the parent
 * should call init_tsc() instead: this allows to minimize the time
 * spent in calibrating.  Each call to init_tsc() involves sleeping
 * for about a second.
 */
void
init_tsc(void)
{
	struct timeval tv_before, tv_after;
	u_int64_t tsc_before, tsc_after;

	gettimeofday(&tv_before, NULL);
	tsc_before = read_tsc();
	sleep(CALIBRATION_INTERVAL);
	gettimeofday(&tv_after, NULL);
	tsc_after = read_tsc();
	tps = (tsc_after - tsc_before)*1000000ULL /
		((tv_after.tv_sec - tv_before.tv_sec)*1000000ULL +
		 (tv_after.tv_usec - tv_before.tv_usec));
	/* We start the epoch at the end of the calibration period. */
	tae = tsc_after;
	/* EPOCH_OFFSET<<32 is the number of fractional seconds
	   between NTP and Unix epochs. */
	epoch = (EPOCH_OFFSET<<32) + (((u_int64_t)tv_after.tv_sec)<<32) +
		(((u_int64_t)tv_after.tv_usec)<<32)/1000000;
}

/*
 * Return the calibrated number of ticks per second.  Must be called
 * after calibration, which is done with init_tsc().  (This function
 * exposes tps in a modular way without a global external variable;
 * global variable in this file is bad enough.)
 */
u_int64_t
get_tps(void)
{
	return tps;
}

/* 
 * Convert a TSC register value into an NTP format (RFC 1305)
 * timestamp.  The ntp pointer must resolve to already allocated
 * memory (8 bytes) that will contain the result of the conversion.
 * NTP format is 4 octets of unsigned integer number of whole seconds
 * since NTP epoch, followed by 4 octets of unsigned integer number of
 * fractional seconds (both numbers are in network byte order).
 */
void
tsc2ntp(u_int64_t tsc, char *ntp)
{
	u_int32_t s, fs;	/* Seconds; fractional seconds. */
	u_int64_t t;		/* Time to our epoch, fractional seconds. */
	u_int64_t t_ntp;	/* Time since NTP epoch, fractional seconds. */
	int after_epoch;	/* Boolean: Is tsc greater than tae? */

	if (tsc > tae) {
		tsc = tsc - tae;	/* Relative to the software epoch. */
		after_epoch = 1;
	} else {
		tsc = tae - tsc;
		after_epoch = 0;
	}
	/*
	 * Compute t, the absolute value of distance to software
	 * epoch.  This is the only operation in tsc2ntp() that loses
	 * precision.
	 */
#if 0
	/* 
	 * This method is more precise, but it places a limit on how
	 * large the offset can be.  With a 2-GHz CPU, offset can't
	 * exceed 2.147 seconds.  Not very usable...
	 */
	assert(tsc < 1ULL<<32);	/* Must not exceed 64 bits after shift. */
	t = (tsc<<32)/tps;
#else
	t = (double)tsc*(double)(1ULL<<32)/(double)tps;
#endif
	if (after_epoch)
		t_ntp = epoch + t;	/* Time relative to NTP epoch. */
	else
		t_ntp = epoch - t;
	s = t_ntp>>32;			/* Most significant 32 bits of t. */
	fs = t_ntp & ((1ULL<<32)-1);	/* Least significant 32 bits t. */
	assert(t_ntp == (((u_int64_t)s)<<32) + fs);
#ifdef DEBUG
	fprintf(stderr, "tsc2ntp:\ttsc=%llu\tt=%llu\ts=%u\tfs=%u\n",
		tsc, t, s, fs);
#endif
	s = htonl(s);
	fs = htonl(fs);
	memcpy(ntp, &s, sizeof(s));
	memcpy(ntp+sizeof(s), &fs, sizeof(fs));
}

/*
 * Convert 8-byte NTP format timestamp into values of local TSC
 * register.  Return the result of the conversion.  The counterpart to
 * tsc2ntp().
 */
u_int64_t
ntp2tsc(const char *ntp)
{
	u_int32_t s, fs;	/* Seconds, fractional seconds. */
	u_int64_t t;		/* Time since epoch in fractional seconds. */

	/* 
	 * We go through the funny dance of assembling the 64-bit
	 * value of t from its halves because of the lack of a
	 * standard ntohll() function.  Otherwise, we'd memcpy(&t,
	 * ntp, sizeof(t), fix it up with t = ntohll(t), and proceed
	 * with subtracting the value of epoch without ever needing
	 * the halves.
	 *
	 * Note that simple t = ntohll(*((u_int64 *)ntp)) would not
	 * work because the buffer ntp might or might not be aligned
	 * properly.  This is also the reason we have to do memcpy()
	 * on the halves below, rather than
	 * s=ntohl(*((u_int32_t*)ntp)).  I could have required the
	 * buffer to be aligned as part of the interface, but that
	 * would, likely, simply move the copying to the callers of
	 * ntp2tsc().
	 */
	memcpy(&s, ntp, sizeof(s));
	memcpy(&fs, ntp+sizeof(s), sizeof(fs));
	s = ntohl(s);
	fs = ntohl(fs);
#ifdef DEBUG
	fprintf(stderr, "ntp2tsc:\ts=%u\tfs=%u\n", s, fs);
#endif
	t = (((u_int64_t)s)<<32) + fs;
	/* The only place precision is lost in this conversion routine
	   is the right shifts below. */
	if (t > epoch)
		return tae + ((tps*(t-epoch))>>32);
	else
		return tae - ((tps*(epoch-t))>>32);
}

#ifdef DEBUG
/* For testing purposes only. */
int
main(void)
{
	u_int64_t t, t1, t2;
	char ntp[8], ntp1[8];

	init_tsc();
	fprintf(stderr, "tae=%llu\ttps=%llu\n", tae, tps);

	t = read_tsc();
	fprintf(stderr, "calling tsc2ntp(t, ntp)\n");
	tsc2ntp(t, ntp);
	t1 = ntp2tsc(ntp);
	fprintf(stderr, "calling tsc2ntp(t1, ntp1)\n");
	tsc2ntp(t1, ntp1);
	t2 = ntp2tsc(ntp1);
	fprintf(stderr,
		"t=%llu\tt1=%llu\ts(t1-t)=%c%lf\nt2=%llu\ts(t2-t1)=%c%lf\n",
		t, t1, t1>=t? '+': '-', ((double)(max(t1,t)-min(t1,t)))/tps,
		t2, t2>=t1? '+': '-', ((double)(max(t2,t1)-min(t2,t1)))/tps);
	exit(0);
}
#endif /* DEBUG */
