/*
 * util.c -- common utility routines for thrulay.
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 * 
 * Copyright 2003, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "util.h"
#include "rcs.h"

RCS_ID("@(#) $Id: util.c,v 1.11 2005/01/12 02:38:16 shalunov Exp $")

void
error(int errcode, const char *msg)
{
	const char *prefix;
	int fatal = 1;

	if (errcode == ERR_FATAL) {
		prefix = "fatal";
	} else if (errcode == ERR_WARNING) {
		prefix = "warning";
		fatal = 0;
	} else {
		prefix = "UNKNOWN ERROR TYPE";
	}
	fprintf(stderr, "%s: %s\n", prefix, msg);
	if (fatal)
		exit(1);
}

ssize_t
read_exactly(int d, void *buf, size_t nbytes)
{
	ssize_t rc = 0;
	size_t bytes_read = 0;
	while (bytes_read < nbytes &&
	       (rc = read(d, (char *)buf+bytes_read,
			  nbytes-bytes_read)) > 0)
		bytes_read += rc;
	return rc == -1? rc: (ssize_t) bytes_read;
}

ssize_t
write_exactly(int d, const void *buf, size_t nbytes)
{
	ssize_t rc = 0;
	size_t bytes_written = 0;
	while (bytes_written < nbytes && 
	       (rc = write(d, (const char *)buf+bytes_written,
			   nbytes-bytes_written)) > 0)
		bytes_written += rc;
	return rc == -1? rc: (ssize_t) bytes_written;
}

double
time_diff(const struct timeval *tv1, const struct timeval *tv2)
{
	return (double) (tv2->tv_sec - tv1->tv_sec)
		+ (double) (tv2->tv_usec - tv1->tv_usec)/1000000.0;
}

/* Set window size for file descriptor FD (which must point to a
   socket) to WINDOW for given direction DIRECTION (typically
   SO_SNDBUF or SO_RCVBUF).  Try hard.  Return actual window size. */
int
set_window_size_directed(int fd, int window, int direction)
{
	int rc, try, w;
	unsigned int optlen = sizeof w;

	rc = getsockopt(fd, SOL_SOCKET, direction, (char *)&w, &optlen);
	if (rc == -1)
		return -1;
	if (window <= 0)
		return w;

	try = window;
	do {
		rc = setsockopt(fd, SOL_SOCKET, direction,
				(char *)&try, optlen);
		try *= 7;
		try /= 8;
	} while (try > w && rc == -1);

	rc = getsockopt(fd, SOL_SOCKET, direction, (char *)&w, &optlen);
	if (rc == -1)
		return -1;
	else
		return w;
}

/* Set window size for file descriptor FD (which must point to a
   socket) to WINDOW.  Try hard.  Return actual window size. */
int
set_window_size(int fd, int window)
{
	int send, receive;

	send = set_window_size_directed(fd, window, SO_SNDBUF);
	receive = set_window_size_directed(fd, window, SO_RCVBUF);
	return send < receive? send: receive;
}

